/* SPDX-License-Identifier: BSD-3-Clause
 * Copyright(c) 2020 Realtek Technologies Co., Ltd
 */

#ifndef _R8168_ETHDEV_H_
#define _R8168_ETHDEV_H_

#include <stdint.h>

#include <rte_ethdev.h>
#include <rte_ethdev_core.h>

#include "r8168_base.h"

struct rtl_hw;

struct rtl_hw_ops {
        //void (*hw_init)(struct rtl_hw *hw);
        void (*hw_config)(struct rtl_hw *hw);
        void (*hw_init_rxcfg)(struct rtl_hw *hw);
        void (*hw_ephy_config)(struct rtl_hw *hw);
        void (*hw_phy_config)(struct rtl_hw *hw);
        void (*hw_mac_mcu_config)(struct rtl_hw *hw);
        void (*hw_phy_mcu_config)(struct rtl_hw *hw);

};

struct rtl_hw {
        u8 *mmio_addr;
        struct rtl_hw_ops hw_ops;
        struct rtl8168_counters *tally_vaddr;
        uint64_t tally_paddr;
        u32 mcfg;
        u32 mtu;
        u32 rx_buf_sz;

        u64 rx_idle_cnt; //debug
        u64 last_rx_packets;

        u8 mac_addr[MAC_ADDR_LEN];
        u8 adapter_stopped;
        u16 cur_page;

        u16 sw_ram_code_ver;
        u16 hw_ram_code_ver;

        u16 phy_reg_anlpar;
        u8 efuse_ver;
        u8 HwSuppGigaForceMode;

        u8  autoneg;
        u8  duplex;
        u32 speed;
        u32 advertising;

        u8 HwSuppPhyOcpVer;
        u8 HwIcVerUnknown;
        u8 HwSuppCheckPhyDisableModeVer;
        u8 HwHasWrRamCodeToMicroP;

        int phy_auto_nego_reg;
        int phy_1000_ctrl_reg;
};

struct rtl_sw_stats {
        u64	tx_packets;
        u64	tx_bytes;
        u64	tx_errors;
        u64	rx_packets;
        u64	rx_bytes;
        u64	rx_errors;
};

struct rtl_adapter {
        struct rtl_hw 		hw;
        struct rtl_sw_stats	sw_stats;
};

#define RTL_DEV_PRIVATE(eth_dev) \
	((struct rtl_adapter *)((eth_dev)->data->dev_private))

int rtl_rx_init(struct rte_eth_dev *dev);
int rtl_tx_init(struct rte_eth_dev *dev);

uint16_t rtl_xmit_pkts(void *txq, struct rte_mbuf **tx_pkts, uint16_t nb_pkts);
uint16_t rtl_recv_pkts(void *rxq, struct rte_mbuf **rx_pkts, uint16_t nb_pkts);

int rtl_start_queues(struct rte_eth_dev *dev);
int rtl_stop_queues(struct rte_eth_dev *dev);
void rtl_free_queues(struct rte_eth_dev *dev);

int rtl_rx_queue_start(struct rte_eth_dev *dev, uint16_t queue_idx);
int rtl_rx_queue_stop(struct rte_eth_dev *dev, uint16_t queue_idx);

int rtl_tx_queue_start(struct rte_eth_dev *dev, uint16_t queue_idx);
int rtl_tx_queue_stop(struct rte_eth_dev *dev, uint16_t queue_idx);

void rtl_tx_queue_release(void *txq);
void rtl_rx_queue_release(void *rxq);

int rtl_tx_done_cleanup(void *tx_queue, uint32_t free_cnt);

int rtl_tx_queue_setup(struct rte_eth_dev *dev, uint16_t queue_idx,
                       uint16_t nb_tx_desc, unsigned int socket_id,
                       const struct rte_eth_txconf *tx_conf);

int rtl_rx_queue_setup(struct rte_eth_dev *dev, uint16_t queue_idx,
                       uint16_t nb_rx_desc, unsigned int socket_id,
                       const struct rte_eth_rxconf *rx_conf,
                       struct rte_mempool *mb_pool);

void rtl_rxq_info_get(struct rte_eth_dev *dev, uint16_t queue_idx,
                      struct rte_eth_rxq_info *qinfo);

void rtl_txq_info_get(struct rte_eth_dev *dev, uint16_t queue_idx,
                      struct rte_eth_txq_info *qinfo);

uint64_t rtl_get_tx_port_offloads(struct rte_eth_dev *dev);
uint64_t rtl_get_rx_port_offloads(struct rte_eth_dev *dev);

void rtl_dump_mac_reg(struct rtl_hw *hw);

#endif /* _R8168_ETHDEV_H_ */
